#ifndef __CONCEPTINTERPRETER_H
#define __CONCEPTINTERPRETER_H

#define USE_JIT_TRACE

#include "Optimizer.h"
#include "Debugger.h"
#include "Array.h"
#include "ConceptPools.h"

#define CONCEPT_STRING(VARIABLE)    (VARIABLE->CLASS_DATA ? (*((AnsiString *)VARIABLE->CLASS_DATA)) : *((AnsiString *)(VARIABLE->CLASS_DATA=new AnsiString())))
//---------------------------------------------------------
#define FREE_VARIABLE(VARIABLE)                                                                \
    VARIABLE->LINKS--;                                                                         \
    if (VARIABLE->LINKS < 1) {                                                                 \
        if (VARIABLE->CLASS_DATA) {                                                            \
            if (VARIABLE->TYPE == VARIABLE_STRING) {                                           \
                delete (AnsiString *)VARIABLE->CLASS_DATA;                                     \
            } else                                                                             \
            if ((VARIABLE->TYPE == VARIABLE_CLASS) || (VARIABLE->TYPE == VARIABLE_DELEGATE)) { \
                if (!--((CompiledClass *)VARIABLE->CLASS_DATA)->LINKS) {                       \
                    delete (CompiledClass *)VARIABLE->CLASS_DATA; }                            \
            } else                                                                             \
            if (VARIABLE->TYPE == VARIABLE_ARRAY) {                                            \
                if (!--((Array *)VARIABLE->CLASS_DATA)->LINKS) {                               \
                    delete (Array *)VARIABLE->CLASS_DATA; }                                    \
            }                                                                                  \
        }                                                                                      \
        VAR_FREE(VARIABLE);                                                                    \
    }

#define FREE_VARIABLE_RESET(VARIABLE, pushed_type)                                             \
    VARIABLE->LINKS--;                                                                         \
    if (VARIABLE->LINKS < 1) {                                                                 \
        if (VARIABLE->CLASS_DATA) {                                                            \
            if (pushed_type == VARIABLE_STRING) {                                              \
                delete (AnsiString *)VARIABLE->CLASS_DATA;                                     \
            } else                                                                             \
            if ((pushed_type == VARIABLE_CLASS) || (pushed_type == VARIABLE_DELEGATE)) {       \
                if (!--((CompiledClass *)VARIABLE->CLASS_DATA)->LINKS) {                       \
                    delete (CompiledClass *)VARIABLE->CLASS_DATA; }                            \
            } else                                                                             \
            if (pushed_type == VARIABLE_ARRAY) {                                               \
                if (!--((Array *)VARIABLE->CLASS_DATA)->LINKS) {                               \
                    delete (Array *)VARIABLE->CLASS_DATA; }                                    \
            }                                                                                  \
        }                                                                                      \
        VAR_FREE(VARIABLE);                                                                    \
    }

#define FAST_FREE_VARIABLE(VARIABLE)                                                           \
    VARIABLE->LINKS--;                                                                         \
    if (VARIABLE->LINKS < 1) {                                                                 \
        if (VARIABLE->CLASS_DATA) {                                                            \
            if (VARIABLE->TYPE == VARIABLE_STRING) {                                           \
                delete (AnsiString *)VARIABLE->CLASS_DATA;                                     \
            } else                                                                             \
            if ((VARIABLE->TYPE == VARIABLE_CLASS) || (VARIABLE->TYPE == VARIABLE_DELEGATE)) { \
                if (!--((CompiledClass *)VARIABLE->CLASS_DATA)->LINKS) {                       \
                    delete (CompiledClass *)VARIABLE->CLASS_DATA; }                            \
            } else                                                                             \
            if (VARIABLE->TYPE == VARIABLE_ARRAY) {                                            \
                if (!--((Array *)VARIABLE->CLASS_DATA)->LINKS) {                               \
                    delete (Array *)VARIABLE->CLASS_DATA; }                                    \
            }                                                                                  \
        }                                                                                      \
    }

#define FREE_VARIABLE_NO_OBJECTS(VARIABLE) \
    VARIABLE->LINKS--;                     \
    if (VARIABLE->LINKS < 1) {             \
        if ((VARIABLE->TYPE == VARIABLE_STRING) && (VARIABLE->CLASS_DATA)) \
            delete (AnsiString *)VARIABLE->CLASS_DATA;                     \
        VAR_FREE(VARIABLE);                                                \
    }
//---------------------------------------------------------
#ifdef SIMPLE_MULTI_THREADING
    #define STACK(_cls, _mem, _caller_line)                                                                          \
        if (PIF->ThreadsCount <= 1) {                                                                                \
            INTERNAL_LOCK(PIF)                                                                                       \
            if (PIF->CALL_STACK_COUNT < MAX_STACK_LIST) {                                             \
                PIF->CALL_STACK_COUNT++;                                                              \
                if (PIF->CALL_STACK_COUNT > 1) {                                                      \
                    PIF->CALL_STACK [PIF->CALL_STACK_COUNT - 2].line = _caller_line; } \
            }                                                                                                        \
            SCStack *SCtarget = &PIF->CALL_STACK [PIF->CALL_STACK_COUNT - 1];          \
            SCtarget->cls  = &_cls;                                                                                  \
            SCtarget->mem  = _mem;                                                                                   \
            SCtarget->line = 0;                                                                                      \
            INTERNAL_UNLOCK(PIF)                                                                                     \
        }

    #define UNSTACK    if (PIF->ThreadsCount <= 1) { INTERNAL_LOCK(PIF) if (PIF->CALL_STACK_COUNT > 0) { PIF->CALL_STACK_COUNT--; } INTERNAL_UNLOCK(PIF) }
#else
    #define STACK(_cls, _mem, _caller_line)                                                                          \
        {                                                                                                            \
            if (PIF->CALL_STACK_COUNT < MAX_STACK_LIST) {                                             \
                PIF->CALL_STACK_COUNT++;                                                              \
                if (PIF->CALL_STACK_COUNT > 1) {                                                      \
                    PIF->CALL_STACK [PIF->CALL_STACK_COUNT - 2].line = _caller_line; } \
            }                                                                                                        \
            SCStack *SCtarget = &PIF->CALL_STACK [PIF->CALL_STACK_COUNT - 1];          \
            SCtarget->cls  = &_cls;                                                                                  \
            SCtarget->mem  = _mem;                                                                                   \
            SCtarget->line = 0;                                                                                      \
        }

    #define UNSTACK    if (PIF->CALL_STACK_COUNT > 0) { PIF->CALL_STACK_COUNT--; }
#endif

class ConceptInterpreter {
    friend class ClassMember;
    friend class ClassCode;
    friend void *GetStaticData();

    friend void SetStaticData(void *);

private:
    Optimizer *OPT;
    //STATIC_FLAG AnsiList     *ClassList;
    INTEGER      LocalClassID;
    ClassMember  *OWNER;
    ParamList *OPERATOR_PARAM;

#ifdef USE_JIT_TRACE
    char callcount;
    void **jittracecode;
    void AnalizeInstructionPath();
#endif
public:
    POOLED(ConceptInterpreter)
    static void AcknoledgeRunTimeError(PIFAlizator *PIF, AnsiException *Exc);

    ConceptInterpreter(Optimizer *O, INTEGER LocalClsID, ClassMember *Owner);
    VariableDATA **CreateEnviroment(PIFAlizator *PIF, VariableDATA *Sender, ParamList *FORMAL_PARAM, VariableDATA **SenderCTX);
    void DestroyEnviroment(PIFAlizator *PIF, VariableDATA **LOCAL_CONTEXT, char static_call_main = 0);
    VariableDATA *Interpret(PIFAlizator *PIF, VariableDATA **LOCAL_CONTEXT, intptr_t ClassID, VariableDATA *&THROW_DATA);
    static int StacklessInterpret(PIFAlizator *PIF, GreenThreadCycle *GREEN);

    ~ConceptInterpreter(void);
};

#endif // __CONCEPTINTERPRETER_H
