//#define DEBUG
//#define TIMMING
#ifdef _WIN32
 #include <windows.h>
#endif

#include "SHManager.h"
#include "AnsiString.h"
#include "PIFAlizator.h"
#include "library.h"
#include "Debugger.h"
#include "ConceptTypes.h"
#include "ConceptInterpreter.h"

#include <sys/types.h>
#include <sys/stat.h>
#ifdef _WIN32
 #include <io.h>

 #include <sys/utime.h>
#else
 #include <unistd.h>
 #include <utime.h>
#endif
#include <stdio.h>
#include <string>
#include <stdlib.h>

#ifdef TIMMING
 #include <time.h>
#endif

POOLED_IMPLEMENTATION(tsVariableDESCRIPTOR);
POOLED_IMPLEMENTATION(tsVariableDATA);
POOLED_IMPLEMENTATION(tsUndefinedMember)
POOLED_IMPLEMENTATION(tsAnalizerElement)
POOLED_IMPLEMENTATION(tsSHORT_AnalizerElement)
POOLED_IMPLEMENTATION(tsPartialAnalizerElement)
POOLED_IMPLEMENTATION(tsOptimizedElement)
POOLED_IMPLEMENTATION(tsParamList)

THREADED_FLAG NOTIFY_PARENT RemoteNotifyParent = 0;
THREADED_FLAG void          *cache_buffer      = 0;

struct Container {
    void *R1;
    void *R2;
    void *R3;
    char is_cgi;
};

static int last_result = 0;
static int cached_direct_pipe = -1;

void *GetPOOLContext(void *refVAR) {
    if ((!refVAR) || (((VariableDATA *)refVAR)->flags < 0))
        return 0;

    VARPool *CURRENT = (VARPool *)((((uintptr_t)refVAR) - sizeof(VariableDATA) * (((VariableDATA *)refVAR)->flags) - POOOL_OFFSET(VARPool, POOL)));
    if (CURRENT)
       return CURRENT->PIF;

    return 0;
}

void VarClean(PIFAlizator *pif) {
    VARPool *CURRENT = pif->POOL;
    VARPool *NEXT = (VARPool *)CURRENT->NEXT;
    int iterations = 0;
    while (NEXT) {
        // skip first block
        CURRENT = NEXT;
		NEXT = (VARPool *)CURRENT->NEXT;
        if (CURRENT->POOL_VARS == POOL_BLOCK_SIZE) {
		    VARPool *PREV  = (VARPool *)CURRENT->PREV;

            PREV->NEXT = NEXT;
            if (NEXT)
                NEXT->PREV = PREV;
            pif->CACHEDPOOL = NULL;
            ((PIFAlizator *)CURRENT->PIF)->free_vars -= POOL_BLOCK_SIZE;

            if (((PIFAlizator *)CURRENT->PIF)->free_vars <= POOL_BLOCK_SIZE)
                NEXT = NULL;
            FAST_FREE(CURRENT);
            iterations = 0;
        } else {
            iterations ++;
            if (iterations == 5)
                break;
        }
    }
}

void FreeMultipleVars(void **refVARs, int count) {
    VariableDATA **CONTEXT = (VariableDATA **)refVARs;
    VARPool *CURRENT = NULL;
    PIFAlizator *pif = NULL;
    for (int i = count - 1; i >= 0; i--) {
        VariableDATA *refVAR = CONTEXT[i];
        if ((refVAR) && (refVAR->LINKS <= 0)) {
            if (refVAR->flags >= 0) {
        	    CURRENT = (VARPool *)(((uintptr_t)refVAR) - sizeof(VariableDATA) * (((VariableDATA *)refVAR)->flags) - POOOL_OFFSET(VARPool, POOL));
                if (((VariableDATA *)refVAR)->flags < CURRENT->FIRST_VAR)
                    CURRENT->FIRST_VAR = ((VariableDATA *)refVAR)->flags;
                pif = (PIFAlizator *)CURRENT->PIF;
                pif->free_vars++;
		        ((VariableDATA *)refVAR)->flags = -1;
		        CURRENT->POOL_VARS ++;
            } else
            if (refVAR->flags == -2) {
                // module allocated!
                FAST_FREE(refVAR);
            }
        }
    }
    if (pif) {
        if (pif->free_vars > BLOCK_SIZE * 10)
            VarClean(pif);
        else
            pif->CACHEDPOOL = CURRENT;
    }
}

void FreeVAR(void *refVAR) {
    if (!refVAR)
        return;

    if (((VariableDATA *)refVAR)->flags == -1)
        return;

    if (((VariableDATA *)refVAR)->flags == -2) {
        // not a pooled variable (module-created variable) for backwards compatibility
        FAST_FREE(refVAR);
        return;
    }

	VARPool *CURRENT = (VARPool *)(((uintptr_t)refVAR) - sizeof(VariableDATA) * (((VariableDATA *)refVAR)->flags) - POOOL_OFFSET(VARPool, POOL));
	//if (CURRENT) {
        ((PIFAlizator *)CURRENT->PIF)->free_vars++;
        if (((VariableDATA *)refVAR)->flags < CURRENT->FIRST_VAR)
            CURRENT->FIRST_VAR = ((VariableDATA *)refVAR)->flags;
		((VariableDATA *)refVAR)->flags = -1;
		CURRENT->POOL_VARS ++;
		VARPool *PREV  = (VARPool *)CURRENT->PREV;
		VARPool *NEXT = (VARPool *)CURRENT->NEXT;
		if ((CURRENT->POOL_VARS == POOL_BLOCK_SIZE) && (PREV)) {
			PREV->NEXT = NEXT;
            if (NEXT)
                NEXT->PREV = CURRENT->PREV;
            ((PIFAlizator *)CURRENT->PIF)->free_vars -= POOL_BLOCK_SIZE;
            if (((PIFAlizator *)CURRENT->PIF)->CACHEDPOOL == CURRENT)
                ((PIFAlizator *)CURRENT->PIF)->CACHEDPOOL = NULL;
			FAST_FREE(CURRENT);
		} else
        if ((CURRENT!=((PIFAlizator *)CURRENT->PIF)->POOL) && (CURRENT->POOL_VARS > ((PIFAlizator *)CURRENT->PIF)->POOL->POOL_VARS)) {
            ((PIFAlizator *)CURRENT->PIF)->CACHEDPOOL = CURRENT;
            if (NEXT)
                NEXT->PREV = CURRENT->PREV;
            PREV->NEXT = NEXT;

            CURRENT->NEXT = ((PIFAlizator *)CURRENT->PIF)->POOL;
            CURRENT->PREV = NULL;
            ((PIFAlizator *)CURRENT->PIF)->POOL->PREV = CURRENT;
            ((PIFAlizator *)CURRENT->PIF)->POOL = CURRENT;
        }
	//}
}

void AllocMultipleVars(void **context, void *PIF, int count, int offset) {
    if (!count)
        return;

    VARPool *POOL = ((PIFAlizator *)PIF)->POOL;
	VARPool *NEXT_POOL;
	int i;
	if (!POOL) {
		POOL = (VARPool *)FAST_MALLOC(sizeof(VARPool));
		POOL->POOL_VARS = POOL_BLOCK_SIZE;
        POOL->FIRST_VAR = 0;
		POOL->NEXT = NULL;
		POOL->PREV = NULL;
        POOL->PIF = PIF;
		for (i=0;i<POOL_BLOCK_SIZE;i++)
			POOL->POOL[i].flags = -1;
        ((PIFAlizator *)PIF)->POOL = POOL;
        ((PIFAlizator *)PIF)->free_vars+=POOL_BLOCK_SIZE;
	}
    
    NEXT_POOL = POOL;
    if (((PIFAlizator *)PIF)->free_vars) {
        int iterations = 0;
        while ((NEXT_POOL) && (((PIFAlizator *)PIF)->free_vars)) {
	        while (NEXT_POOL->POOL_VARS) {
		        for (i=NEXT_POOL->FIRST_VAR;i<POOL_BLOCK_SIZE;i++) {
			        if (NEXT_POOL->POOL[i].flags == -1) {
				        NEXT_POOL->POOL_VARS--;
				        NEXT_POOL->POOL[i].flags = i;
                        ((PIFAlizator *)PIF)->free_vars--;
                        context[offset++] = &NEXT_POOL->POOL[i];
				        count--;
                        NEXT_POOL->FIRST_VAR = i+1;
                        if (!count)
                            return;
			        }
		        }
            }

            iterations++;
            if (((PIFAlizator *)PIF)->free_vars < POOL_BLOCK_SIZE)
                break;
            // memory is too fragmented ...
            if (iterations > 1000)
                break;
            NEXT_POOL = (VARPool *)NEXT_POOL->NEXT;
        }
    }

    while (count) {
	    NEXT_POOL = (VARPool *)FAST_MALLOC(sizeof(VARPool));
        if (!NEXT_POOL) {
            ((PIFAlizator *)PIF)->out->ClientError(ERR1400 "\n");
            exit(-100);
        }
        if (count < POOL_BLOCK_SIZE) {
            NEXT_POOL->FIRST_VAR = count;
	        NEXT_POOL->POOL_VARS = POOL_BLOCK_SIZE - count;
            ((PIFAlizator *)PIF)->free_vars += NEXT_POOL->POOL_VARS;
        } else {
            NEXT_POOL->FIRST_VAR = POOL_BLOCK_SIZE;
	        NEXT_POOL->POOL_VARS = 0;
        }

        NEXT_POOL->PIF = PIF;
	    for (i=0;i<POOL_BLOCK_SIZE;i++) {
            if (count) {
                count--;
                //((PIFAlizator *)PIF)->free_vars--;
                //NEXT_POOL->POOL_VARS--;
		        NEXT_POOL->POOL[i].flags = i;
                context[offset++] = &NEXT_POOL->POOL[i];
            } else
		        NEXT_POOL->POOL[i].flags = -1;
        }

	    NEXT_POOL->PREV = POOL->PREV;
	    NEXT_POOL->NEXT = POOL;
	    POOL->PREV = NEXT_POOL;
	    ((PIFAlizator *)PIF)->POOL = NEXT_POOL;
        POOL = NEXT_POOL;
    }
}

void *AllocVAR(void *PIF) {
    if (!PIF) {
        // pool variable
        VariableDATA *notPOOLED = (VariableDATA *)FAST_MALLOC(sizeof(VariableDATA));
        if (notPOOLED)
            notPOOLED->flags = -2;
        return notPOOLED;
    }
    VARPool *POOL = ((PIFAlizator *)PIF)->POOL;
	VARPool *NEXT_POOL;
	int i;
	if (!POOL) {
		POOL = (VARPool *)FAST_MALLOC(sizeof(VARPool));
        POOL->FIRST_VAR = 0;
		POOL->POOL_VARS = POOL_BLOCK_SIZE;
		POOL->NEXT = NULL;
		POOL->PREV = NULL;
        POOL->PIF = PIF;
		for (i=0;i<POOL_BLOCK_SIZE;i++)
			POOL->POOL[i].flags = -1;
        ((PIFAlizator *)PIF)->POOL = POOL;
        ((PIFAlizator *)PIF)->free_vars+=POOL_BLOCK_SIZE;
	}
    NEXT_POOL = ((PIFAlizator *)PIF)->CACHEDPOOL ? ((PIFAlizator *)PIF)->CACHEDPOOL : POOL;
    if (((PIFAlizator *)PIF)->free_vars) {
        int iterations = 0;
        while ((NEXT_POOL) && (((PIFAlizator *)PIF)->free_vars)) {
	        if (NEXT_POOL->POOL_VARS) {
		        for (i=NEXT_POOL->FIRST_VAR;i<POOL_BLOCK_SIZE;i++) {
			        if (NEXT_POOL->POOL[i].flags == -1) {
				        NEXT_POOL->POOL_VARS--;
				        NEXT_POOL->POOL[i].flags = i;
                        if (NEXT_POOL->POOL_VARS)
                            NEXT_POOL->FIRST_VAR = i+1;
                        else
                            NEXT_POOL->FIRST_VAR = POOL_BLOCK_SIZE;
                        ((PIFAlizator *)PIF)->free_vars--;
                        if (((PIFAlizator *)PIF)->free_vars)
                            ((PIFAlizator *)PIF)->CACHEDPOOL = NEXT_POOL;
				        return &NEXT_POOL->POOL[i];
			        }
		        }
	        }
            iterations++;
            if (((PIFAlizator *)PIF)->free_vars < POOL_BLOCK_SIZE)
                break;
            // memory is too fragmented ...
            if (iterations > 200)
                break;
            NEXT_POOL = (VARPool *)NEXT_POOL->NEXT;
        }
    }
    ((PIFAlizator *)PIF)->CACHEDPOOL = NULL;
	NEXT_POOL = (VARPool *)FAST_MALLOC(sizeof(VARPool));
    if (!NEXT_POOL) {
        ((PIFAlizator *)PIF)->out->ClientError(ERR1400 "\n");
        exit(-100);
    }
    NEXT_POOL->FIRST_VAR = 1;
    NEXT_POOL->PIF = PIF;
	NEXT_POOL->POOL_VARS = POOL_BLOCK_SIZE - 1;
    ((PIFAlizator *)PIF)->free_vars += NEXT_POOL->POOL_VARS;
	for (i=1;i<POOL_BLOCK_SIZE;i++)
		NEXT_POOL->POOL[i].flags = -1;
	NEXT_POOL->PREV = POOL->PREV;
	NEXT_POOL->NEXT = POOL;
	POOL->PREV = NEXT_POOL;
    ((PIFAlizator *)PIF)->CACHEDPOOL = NEXT_POOL;
	((PIFAlizator *)PIF)->POOL = NEXT_POOL;
	NEXT_POOL->POOL[0].flags = 0;
    //((PIFAlizator *)PIF)->free_vars--;
	return &NEXT_POOL->POOL[0];
}
/*
void PoolStats(void *PIF) {
    int pools = 0;
    int vars = 0;
    VARPool *NEXT_POOL = ((PIFAlizator *)PIF)->POOL;
    while (NEXT_POOL) {
	    if (NEXT_POOL->POOL_VARS) {
		    for (int i=0;i<BLOCK_SIZE;i++) {
			    if (NEXT_POOL->POOL[i].flags != -1) {
				    vars++;
			    }
		    }
	    }
        NEXT_POOL = (VARPool *)NEXT_POOL->NEXT;
        pools++;
    }
    fprintf(stderr, "POOLS %i, VARS: %i\n", pools, vars);
}
*/
CachedFILE *cached_fopen(const char *filename, const char *mode) {
    CachedFILE *fin = 0;
    FILE       *in  = fopen(filename, mode);

    if (in) {
        fin = (CachedFILE *)malloc(sizeof(CachedFILE));
        if (fin) {
            fin->buf_fill = 0;
            fin->buf_pos  = 0;
            fin->in       = in;
            return(fin);
        } else {
            fclose(in);
        }
    }
    return(0);
}

int cached_fclose(CachedFILE *in) {
    int res = 0;

    if (in) {
        res = fclose(in->in);
        free(in);
    }
    return(res);
}

int cached_fseek(CachedFILE *in, intptr_t offset, int origin) {
    switch (origin) {
        case SEEK_CUR:
            {
                int delta = in->buf_fill - in->buf_pos;
                if (offset < delta) {
                    in->buf_pos += offset;
                } else {
                    offset -= delta;
                    if (offset) {
                        fseek(in->in, offset, SEEK_CUR);
                    }
                    in->buf_fill = 0;
                    in->buf_pos  = 0;
                }
            }
            return(0);
            break;

        case SEEK_SET:
        case SEEK_END:
        default:
            in->buf_fill = 0;
            in->buf_pos  = 0;
            return(fseek(in->in, offset, origin));
            break;
    }
    return(-1);
}

int cached_fread(void *buf, size_t size, size_t len, CachedFILE *in) {
    int total_size = size * len;
    int total      = 0;

    if ((!in->buf_fill) || (in->buf_pos >= in->buf_fill)) {
        in->buf_fill = fread(in->sbuf, 1, sizeof(in->sbuf), in->in);
        in->buf_pos  = 0;
    }
    if (in->buf_fill > 0) {
        int remaining = in->buf_fill - in->buf_pos;
        if (total_size <= remaining) {
            memcpy(buf, in->sbuf + in->buf_pos, total_size);
            in->buf_pos += total_size;
            total        = total_size;
        } else {
            if (remaining) {
                memcpy(buf, in->sbuf + in->buf_pos, remaining);
                total      += remaining;
                total_size -= remaining;
            }
            do {
                in->buf_fill = fread(in->sbuf, 1, sizeof(in->sbuf), in->in);
                in->buf_pos  = 0;
                if (in->buf_fill > 0) {
                    if (total_size <= in->buf_fill) {
                        memcpy(((char *)buf) + total, in->sbuf, total_size);
                        in->buf_pos = total_size;
                        total      += total_size;
                        break;
                    } else {
                        memcpy(((char *)buf) + total, in->sbuf, in->buf_fill);
                        total      += in->buf_fill; 
                        total_size -= in->buf_fill;
                        in->buf_pos = in->buf_fill;
                    }
                }
            } while (in->buf_fill > 0);
        }
    }
    return(total / size);
}

int NotifyParent(int pipe_out, int apid, int msg_id, AnsiString DATA) {
    if (RemoteNotifyParent) {
        return(RemoteNotifyParent(pipe_out, apid, msg_id, DATA.Length(), DATA.c_str()));
    }

    if ((apid <= 0) || (pipe_out <= 0)) {
        return(0);
    }

    int len = DATA.Length();
    char *ptr = DATA.c_str();
    //while (len > 0) {
        //int len2 = len > 0x200 ? 0x200 : len;
        int params[3];
        params[0] = apid;
        params[1] = msg_id;
        params[2] = len;

        write(pipe_out, params, sizeof(int) * 3);
        if (len) {
            while (len > 0) {
                int written = write(pipe_out, ptr, len);
                if (written > 0) {
                    len -= written;
                    ptr += written;
                } else
                    break;
            }
        }
        //len -= len2;
        //ptr += len2;
    //}
    return(1);
}

void DeturnatedPrint(void *PIF, char *text, int len, void *userdata) {
    AnsiString data;

    if (len > -1) {
        data.LoadBuffer(text, len);
    } else {
        data = text;
    }
    if (PIF)
        NotifyParent(((PIFAlizator *)PIF)->pipe_write, ((PIFAlizator *)PIF)->parent_apid, -10, data);
}

CONCEPT_DLL_API SetNotifyParent(NOTIFY_PARENT np) {
    RemoteNotifyParent = np;
    return(0);
}

CONCEPT_DLL_API SetDirectPipe(int direct_pipe) {
    //PIFAlizator::direct_pipe=direct_pipe;
    cached_direct_pipe = direct_pipe;
    return(0);
}

CONCEPT_DLL_API SetArguments(int argc, char **argv) {
    PIFAlizator::argc = argc;
    PIFAlizator::argv = argv;
    return(0);
}

struct ThreadData {
    NOTIFY_PARENT     RemoteNotifyParent;
    int               argc;
    char              **argv;
    //AnsiList          *ClassList;
};

void *GetStaticData() {
#ifndef SINGLE_PROCESS_DLL
    ThreadData *td = new ThreadData();
    td->RemoteNotifyParent = RemoteNotifyParent;
    td->argc = PIFAlizator::argc;
    td->argv = PIFAlizator::argv;

    //td->ClassList          = ConceptInterpreter::ClassList;
    return(td);
#else
    return(0);
#endif
}

void SetStaticData(void *data) {
#ifndef SINGLE_PROCESS_DLL
    ThreadData *td = (ThreadData *)data;
    RemoteNotifyParent = td->RemoteNotifyParent;
    PIFAlizator::argc  = td->argc;
    PIFAlizator::argv  = td->argv;

    //ConceptInterpreter::ClassList          = td->ClassList;
    delete td;
#endif
}

CONCEPT_DLL_API Concept_CachedInit(char *filename) {
    struct stat buf;
    char lpBuffer[8192];
    AnsiString  accel_filename(filename);
    accel_filename += (char *)DEFAULT_BIN_EXTENSION;

#ifdef _WIN32
    AnsiString key;
    lpBuffer[8191]=0;
    GetCurrentDirectory(8191, lpBuffer);
    int i=0;
    while (lpBuffer[i]) {
        if (lpBuffer[i]=='\\')
            lpBuffer[i]='/';
        i++;
    }
    key+=lpBuffer;
    key+=(char *)"/";
    key+=filename;
#else
    AnsiString key;
    lpBuffer[8191]=0;
    getcwd(lpBuffer, 8191);
    key+=lpBuffer;
    key+=(char *)"/";
    key+=filename;
#endif
    FILE *in = fopen(accel_filename.c_str(), "rb");
    if (!in)
        return(-2);
    fclose(in);
#ifdef _WIN32
    if (!stat(accel_filename.c_str(), &buf)) {
        key += (char *)":";
        key += AnsiString((intptr_t)buf.st_size);
        key += (char *)":";
        key += AnsiString((intptr_t)buf.st_mtime);
    }
#endif
    cache_buffer = SHPoolOpen(key.c_str());
    SHLockWait(key.c_str());
    if (cache_buffer) {
        return(0xFFFF);
    }
    if (!cache_buffer) {
        if (!SHLock(key.c_str()))
            return -1;
        // 2. create buffer, compute size !
        int size = PIFAlizator::ComputeSharedSize(accel_filename.c_str());
        if (size > 0) {
            // 255 bytes reserved ...
            size += 0xFF;
            cache_buffer = SHPoolAlloc(key.c_str(), size);
            // don't unlock the semaphore !!
            if (cache_buffer) {
                return(size);
            }
        }
        SHUnLock();
    }
    return(-1);
}

CONCEPT_DLL_API Concept_CachedDone(char *filename) {
    if (cache_buffer) {
        SHPoolFree(cache_buffer);
        cache_buffer = 0;
        SHUnLock();
    }
    return(0);
}

CONCEPT_DLL_API Concept_Compile(char *filename, char *inc_dir, char *lib_dir, ForeignPrint fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent) {
    // programm internal form
    AnsiString S;

    if (S.LoadFile(filename)) {
        return(-3);
    }

    SimpleStream SS(fp, sock);

    char is_cgi = 0;
    if (debug == -1) {
        is_cgi = 1;
        debug  = 0;
    }
#ifdef TIMMING
    intptr_t start = clock();
    SS.Print("PIF...");
#endif
    PIFAlizator PIF(inc_dir, lib_dir, &S, &SS, filename, debug, DEBUGGER_TRAP, DEBUGGER_RESERVED, SERVER_PUBLIC_KEY, SERVER_PRIVATE_KEY, CLIENT_PUBLIC_KEY);
    PIF.SetPipe(pipe_in, pipe_out, apid, parent, cached_direct_pipe);
    PIF.Execute();
#ifdef TIMMING
    intptr_t total = clock() - start;
    SS.Print("done\nIn: ");
    SS.Print((double)total);
    SS.Print("\n");
    SS.Print("OPTIMIZE...");
    start = clock();
#endif
    bool build_package = false;
    if (!PIF.ErrorCount()) {
        PIF.Optimize();
        build_package = !PIF.CheckRunable();

        if ((PIF.Errors.Count() == 1) && (build_package)) {
            PIF.Errors.Delete(PIF.Errors.Count() - 1);
        }

#ifdef OPTIMIZE_MEMORY
        if (!PIF.Errors.Count()) {
            SS.Print("Optimising memory usage ...\n");
            PIF.OptimizeMemoryUsage();
        }
#endif
        if (build_package) {
            //PIF.Errors.Clear();
            SS.Print("Building package ...\n");
        }
    }
#ifdef TIMMING
    total = clock() - start;
    SS.Print("done\nIn: ");
    SS.Print((double)total);
    SS.Print("\n");
#endif
    if (!PIF.ErrorCount()) {
        if ((PIF.WarningCount()) && (!is_cgi)) {
            SS.Print(PIF.PRINT_WARNINGS());
            SS.Print("\n\nThere are warnings.\n");
        }
        AnsiString accel_filename = AnsiString(filename);
        accel_filename += build_package ? (char *)DEFAULT_PACK_EXTENSION : (char *)DEFAULT_BIN_EXTENSION;
        PIF.Serialize(accel_filename, build_package);
        // set file times ...
        struct stat buf;
        if (!stat(filename, &buf)) {
            struct utimbuf timebuf;
            timebuf.actime  = buf.st_mtime;
            timebuf.modtime = buf.st_mtime;

            // set the time for the accel file
            utime(accel_filename.c_str(), &timebuf);
        }
    } else {
        SS.Print(PIF.PRINT_ERRORS(is_cgi));
        return(-2);
    }
    return(0);
}

CONCEPT_DLL_API LastResult() {
    return(last_result);
}

CONCEPT_DLL_API Concept_Execute(char *filename, char *inc_dir, char *lib_dir, ForeignPrint fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent, void *userdata) {
    ForeignNotify fp_notify = 0;

    if ((parent > 0) && (pipe_out > 0))
        fp_notify       = DeturnatedPrint;

    SimpleStream SS(fp, sock, fp_notify, userdata);
    AnsiString   S;
    if (S.LoadFile(filename)) {
        NotifyParent(pipe_out, parent, -6, "File not found");
        SS.Print(AnsiString("File '") + AnsiString(filename) + "' is not available\n", -1, -2);
        return(-3);
    }

    char is_cgi = 0;

    if (debug == -2) {
        is_cgi = 1;
        debug  = 1;
    }

    if (debug == -1) {
        is_cgi = 1;
        debug  = 0;
    }
    PIFAlizator PIF(inc_dir, lib_dir, &S, &SS, filename, debug, DEBUGGER_TRAP, DEBUGGER_RESERVED, SERVER_PUBLIC_KEY, SERVER_PRIVATE_KEY, CLIENT_PUBLIC_KEY);
    PIF.SetPipe(pipe_in, pipe_out, apid, parent, cached_direct_pipe);
    NotifyParent(pipe_out, parent, -1, "Unserializing");
#ifdef TIMMING
    int start=clock();
#endif
    if ((!debug) && (!PIF.Unserialize(AnsiString(filename) + DEFAULT_BIN_EXTENSION))) {
        NotifyParent(pipe_out, parent, -1, "Optimizing");
        PIF.Optimize(true);
#ifdef TIMMING
        printf("Total time: %i\n", clock()-start);
#endif
        // in case we have some cache locking
        SHUnLock();
    } else {
        NotifyParent(pipe_out, parent, -1, "Compiling");
        PIF.Execute();
        if (!PIF.ErrorCount()) {
            NotifyParent(pipe_out, parent, -1, "Optimizing");
            PIF.Optimize();
#ifdef OPTIMIZE_MEMORY
            NotifyParent(pipe_out, parent, -1, "Optimizing memory usage");
            PIF.OptimizeMemoryUsage();
#endif
        }
    }
    if (!PIF.ErrorCount()) {
        PIF.CheckRunable();
    }
    if (!PIF.ErrorCount()) {
        if ((PIF.WarningCount()) && (!is_cgi)) {
            SS.Print(PIF.PRINT_WARNINGS());
            SS.Print("\n\nThere are warnings.\n");
            NotifyParent(pipe_out, parent, -5, PIF.SerializeWarningsErrors(1));
        }
        NotifyParent(pipe_out, parent, -4, "Running");
        ClassCode *CC = (ClassCode *)PIF.GetStartingPoint();

        VariableDATA *STATIC_ENTRY = (VariableDATA *)VAR_ALLOC(&PIF);
        STATIC_ENTRY->CLASS_DATA         = 0;
        STATIC_ENTRY->LINKS              = 2;              // never to be deleted
        STATIC_ENTRY->NUMBER_DATA        = 0;
        STATIC_ENTRY->TYPE               = VARIABLE_CLASS; //TYPE_CLASS;
        //STATIC_ENTRY->CALL_SET           = 0;
        STATIC_ENTRY->IS_PROPERTY_RESULT = 0;

        // nu mai e nevoie, pt ca se sterge singur ...
        STATIC_ENTRY->CLASS_DATA = CC->CreateInstance(&PIF, STATIC_ENTRY, 0, 0, 0, 1);
        STATIC_ENTRY->CLASS_DATA = 0;

        VAR_FREE(STATIC_ENTRY);

        NotifyParent(pipe_out, parent, -6, "Done");
        last_result = PIF.last_result;
        if (PIF.ErrorCount()) {
            if (SS.AlreadyPrinted()) {
                SS.Print(PIF.PRINT_ERRORS(is_cgi ? 2 : 0));
            } else {
                SS.Print(PIF.PRINT_ERRORS(is_cgi));
            }
            if (!is_cgi) {
                SS.Print("There are RUN-TIME errors.", -1, -2);
            } else {
                SS.Print("\n\nThere are RUN-TIME errors.\n");
            }
            NotifyParent(pipe_out, parent, -2, PIF.SerializeWarningsErrors());
            return(-1);
        }
    } else {
        if (SS.AlreadyPrinted()) {
            SS.Print(PIF.PRINT_ERRORS(is_cgi ? 2 : 0));
        } else {
            SS.Print(PIF.PRINT_ERRORS(is_cgi));
        }

        if (!is_cgi) {
            SS.Print("There are compile-time errors.", -1, -2);
        }

        NotifyParent(pipe_out, parent, -3, PIF.SerializeWarningsErrors());
        return(-2);
    }
    //PoolStats(&PIF);
    return(0);
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API SetCheckpointFunction(CHECK_POINT cp) {
    PIFAlizator::CheckPoint = cp;
    return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API Concept_Execute2(char *filename, char *inc_dir, char *lib_dir, ForeignPrint fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent, CHECK_POINT cp, void *userdata) {
    PIFAlizator::CheckPoint = cp;
    return(Concept_Execute(filename, inc_dir, lib_dir, fp, sock, debug, DEBUGGER_TRAP, DEBUGGER_RESERVED, SERVER_PUBLIC_KEY, SERVER_PRIVATE_KEY, CLIENT_PUBLIC_KEY, pipe_in, pipe_out, apid, parent, userdata));
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API Concept_GetSocket() {
    return PIFAlizator::refSOCKET;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API Concept_ExecuteBuffer(char *buffer, int len, char *inc_dir, char *lib_dir, ForeignPrint fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent, CHECK_POINT cp, void *result, void *module_list, void *userdata) {
    PIFAlizator::CheckPoint = cp;
    ForeignNotify fp_notify = 0;

    if ((parent > 0) && (pipe_out > 0))
        fp_notify       = DeturnatedPrint;

    SimpleStream SS(fp, sock, fp_notify, userdata);
    AnsiString   S;
    S.LoadBuffer(buffer, len);

    char is_cgi = 0;

    if (debug == -2) {
        is_cgi = 1;
        debug  = 1;
    }

    if (debug == -1) {
        is_cgi = 1;
        debug  = 0;
    }
    PIFAlizator PIF(inc_dir, lib_dir, &S, &SS, "code", debug, DEBUGGER_TRAP, DEBUGGER_RESERVED, SERVER_PUBLIC_KEY, SERVER_PRIVATE_KEY, CLIENT_PUBLIC_KEY, NULL, result, module_list);
    PIF.SetPipe(pipe_in, pipe_out, apid, parent, cached_direct_pipe);
    NotifyParent(pipe_out, parent, -1, "Unserializing");
#ifdef TIMMING
    int start=clock();
#endif

    NotifyParent(pipe_out, parent, -1, "Compiling");
    PIF.Execute();
    if (!PIF.ErrorCount()) {
        NotifyParent(pipe_out, parent, -1, "Optimizing");
        PIF.Optimize();
#ifdef OPTIMIZE_MEMORY
        NotifyParent(pipe_out, parent, -1, "Optimizing memory usage");
        PIF.OptimizeMemoryUsage();
#endif
    }
    if (!PIF.ErrorCount()) {
        PIF.CheckRunable();
    }
    if (!PIF.ErrorCount()) {
        if ((PIF.WarningCount()) && (!is_cgi)) {
            SS.Print(PIF.PRINT_WARNINGS());
            SS.Print("\n\nThere are warnings.\n");
            NotifyParent(pipe_out, parent, -5, PIF.SerializeWarningsErrors(1));
        }
        NotifyParent(pipe_out, parent, -4, "Running");
        ClassCode *CC = (ClassCode *)PIF.GetStartingPoint();

        VariableDATA *STATIC_ENTRY = (VariableDATA *)VAR_ALLOC(&PIF);
        STATIC_ENTRY->CLASS_DATA         = 0;
        STATIC_ENTRY->LINKS              = 2;              // never to be deleted
        STATIC_ENTRY->NUMBER_DATA        = 0;
        STATIC_ENTRY->TYPE               = VARIABLE_CLASS; //TYPE_CLASS;
        //STATIC_ENTRY->CALL_SET           = 0;
        STATIC_ENTRY->IS_PROPERTY_RESULT = 0;

        // nu mai e nevoie, pt ca se sterge singur ...
        STATIC_ENTRY->CLASS_DATA = CC->CreateInstance(&PIF, STATIC_ENTRY, 0, 0, 0, 1);
        STATIC_ENTRY->CLASS_DATA = 0;

        VAR_FREE(STATIC_ENTRY);

        NotifyParent(pipe_out, parent, -6, "Done");
        last_result = PIF.last_result;
        if (PIF.ErrorCount()) {
            if (SS.AlreadyPrinted()) {
                SS.Print(PIF.PRINT_ERRORS(is_cgi ? 2 : 0));
            } else {
                SS.Print(PIF.PRINT_ERRORS(is_cgi));
            }
            if (!is_cgi) {
                SS.Print("There are RUN-TIME errors.", -1, -2);
            } else {
                SS.Print("\n\nThere are RUN-TIME errors.\n");
            }
            NotifyParent(pipe_out, parent, -2, PIF.SerializeWarningsErrors());
            return(-1);
        }
    } else {
        if (SS.AlreadyPrinted()) {
            SS.Print(PIF.PRINT_ERRORS(is_cgi ? 2 : 0));
        } else {
            SS.Print(PIF.PRINT_ERRORS(is_cgi));
        }

        if (!is_cgi) {
            SS.Print("There are compile-time errors.", -1, -2);
        }

        NotifyParent(pipe_out, parent, -3, PIF.SerializeWarningsErrors());
        return(-2);
    }
    return(0);
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API_HANDLER Concept_Execute3_Init(char *filename, char *inc_dir, char *lib_dir, ForeignPrint fp, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent, void *userdata) {
    ForeignNotify fp_notify = 0;

    if ((parent > 0) && (pipe_out > 0))
        fp_notify       = DeturnatedPrint;

    SimpleStream *SS=new SimpleStream(fp, sock, fp_notify, userdata);
    AnsiString   S;
    if (S.LoadFile(filename)) {
        NotifyParent(pipe_out, parent, -6, "File not found");
        SS->Print(AnsiString("File '") + AnsiString(filename) + "' is not available\n", -1, -2);
        return(0);
    }

    char is_cgi = 0;

    if (debug == -2) {
        is_cgi = 1;
        debug  = 1;
    }

    if (debug == -1) {
        is_cgi = 1;
        debug  = 0;
    }
    PIFAlizator *PIF=new PIFAlizator(inc_dir, lib_dir, &S, SS, filename, debug, DEBUGGER_TRAP, DEBUGGER_RESERVED, SERVER_PUBLIC_KEY, SERVER_PRIVATE_KEY, CLIENT_PUBLIC_KEY);
    PIF->SetPipe(pipe_in, pipe_out, apid, parent, cached_direct_pipe);
    NotifyParent(pipe_out, parent, -1, "Unserializing");
    if ((!debug) && (!PIF->Unserialize(AnsiString(filename) + DEFAULT_BIN_EXTENSION))) {
        NotifyParent(pipe_out, parent, -1, "Optimizing");
        PIF->Optimize(true);
        // in case we have some cache locking
        SHUnLock();
    } else {
        NotifyParent(pipe_out, parent, -1, "Compiling");
        PIF->Execute();
        if (!PIF->ErrorCount()) {
            NotifyParent(pipe_out, parent, -1, "Optimizing");
            PIF->Optimize();
#ifdef OPTIMIZE_MEMORY
            NotifyParent(pipe_out, parent, -1, "Optimizing memory usage");
            PIF.OptimizeMemoryUsage();
#endif
        }
    }
    if (!PIF->ErrorCount()) {
        PIF->CheckRunable();
    }
    if (!PIF->ErrorCount()) {
        if ((PIF->WarningCount()) && (!is_cgi)) {
            SS->Print(PIF->PRINT_WARNINGS());
            SS->Print("\n\nThere are warnings.\n");
            NotifyParent(pipe_out, parent, -5, PIF->SerializeWarningsErrors(1));
        }
        NotifyParent(pipe_out, parent, -4, "Running");

        ClassCode *CC = (ClassCode *)PIF->GetStartingPoint();

        Container *res=new Container;
        res->R1=PIF;
        res->R2=SS;
        res->R3=CC;
        res->is_cgi=is_cgi;
        return res;
    } else {
        if (SS->AlreadyPrinted()) {
            SS->Print(PIF->PRINT_ERRORS(is_cgi ? 2 : 0));
        } else {
            SS->Print(PIF->PRINT_ERRORS(is_cgi));
        }

        if (!is_cgi) {
            SS->Print("There are compile-time errors.", -1, -2);
        }

        NotifyParent(pipe_out, parent, -3, PIF->SerializeWarningsErrors());
        delete PIF;
        delete SS;
        return(0);
    }
    return(0);
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API Concept_Execute3_Run(void *PTR, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent) {
    Container *Cptr=(Container *)PTR;

    char is_cgi = 0;

    if (debug == -2) {
        is_cgi = 1;
        debug  = 1;
    }

    if (debug == -1) {
        is_cgi = 1;
        debug  = 0;
    }

    if (!Cptr)
        return -1;
    SimpleStream *SS=(SimpleStream *)Cptr->R2;
    PIFAlizator *PIF=(PIFAlizator *)Cptr->R1;
    ClassCode *CC=(ClassCode *)Cptr->R3;
    SS->Reinit(sock);

    PIF->Adjust(debug, DEBUGGER_TRAP, DEBUGGER_RESERVED, SERVER_PUBLIC_KEY, SERVER_PRIVATE_KEY, CLIENT_PUBLIC_KEY);
    PIF->SetPipe(pipe_in, pipe_out, apid, parent, cached_direct_pipe);

    VariableDATA *STATIC_ENTRY = (VariableDATA *)VAR_ALLOC(PIF);
    STATIC_ENTRY->CLASS_DATA         = 0;
    STATIC_ENTRY->LINKS              = 2;              // never to be deleted
    STATIC_ENTRY->NUMBER_DATA        = 0;
    STATIC_ENTRY->TYPE               = VARIABLE_CLASS; //TYPE_CLASS;
    //STATIC_ENTRY->CALL_SET           = 0;
    STATIC_ENTRY->IS_PROPERTY_RESULT = 0;

    // nu mai e nevoie, pt ca se sterge singur ...
    STATIC_ENTRY->CLASS_DATA = CC->CreateInstance(PIF, STATIC_ENTRY, 0, 0, 0/*, 1*/);
    STATIC_ENTRY->CLASS_DATA = 0;

    VAR_FREE(STATIC_ENTRY);

    NotifyParent(pipe_out, parent, -6, "Done");
    last_result = PIF->last_result;
    if (PIF->ErrorCount()) {
        if (SS->AlreadyPrinted()) {
            SS->Print(PIF->PRINT_ERRORS(Cptr->is_cgi ? 2 : 0));
        } else {
            SS->Print(PIF->PRINT_ERRORS(Cptr->is_cgi));
        }
        if (!Cptr->is_cgi) {
            SS->Print("There are RUN-TIME errors.", -1, -2);
        } else {
            SS->Print("\n\nThere are RUN-TIME errors.\n");
        }

        NotifyParent(pipe_out, parent, -2, PIF->SerializeWarningsErrors());
        PIF->Clear();
        return(0);
    }
    PIF->Clear();
    return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API Concept_Execute3_RunClone(void *PTR, SOCKET sock, int debug, DEBUGGER_CALLBACK DEBUGGER_TRAP, void *DEBUGGER_RESERVED, char *SERVER_PUBLIC_KEY, char *SERVER_PRIVATE_KEY, char *CLIENT_PUBLIC_KEY, int pipe_in, int pipe_out, int apid, int parent, int direct_pipe) {
    Container *Cptr=(Container *)PTR;

    char is_cgi = 0;

    if (debug == -2) {
        is_cgi = 1;
        debug  = 1;
    }

    if (debug == -1) {
        is_cgi = 1;
        debug  = 0;
    }

    if (!Cptr)
        return -1;

    ForeignNotify fp_notify = 0;

    if ((parent > 0) && (pipe_out > 0))
        fp_notify       = DeturnatedPrint;

    SimpleStream *SSParent=(SimpleStream *)Cptr->R2;
    PIFAlizator *PIFParent=(PIFAlizator *)Cptr->R1;
    SimpleStream SS(SSParent->fprint, sock, fp_notify, SSParent->userdata);
    AnsiString S;
    PIFAlizator PIF("", "", &S, &SS, "", debug, DEBUGGER_TRAP, DEBUGGER_RESERVED, SERVER_PUBLIC_KEY, SERVER_PRIVATE_KEY, CLIENT_PUBLIC_KEY, PIFParent);
    PIF.SetPipe(pipe_in, pipe_out, apid, parent, direct_pipe);

    ClassCode *CC=(ClassCode *)Cptr->R3;

    VariableDATA *STATIC_ENTRY = (VariableDATA *)VAR_ALLOC(&PIF);
    STATIC_ENTRY->CLASS_DATA         = 0;
    STATIC_ENTRY->LINKS              = 2;              // never to be deleted
    STATIC_ENTRY->NUMBER_DATA        = 0;
    STATIC_ENTRY->TYPE               = VARIABLE_CLASS;
    //STATIC_ENTRY->CALL_SET           = 0;
    STATIC_ENTRY->IS_PROPERTY_RESULT = 0;

    // nu mai e nevoie, pt ca se sterge singur ...
    STATIC_ENTRY->CLASS_DATA = CC->CreateInstance(&PIF, STATIC_ENTRY, 0, 0, 0, 1/*, 1*/);
    STATIC_ENTRY->CLASS_DATA = 0;

    VAR_FREE(STATIC_ENTRY);

    NotifyParent(pipe_out, parent, -6, "Done");
    last_result = PIF.last_result;
    if (PIF.ErrorCount()) {
        if (SS.AlreadyPrinted()) {
            SS.Print(PIF.PRINT_ERRORS(Cptr->is_cgi ? 2 : 0));
        } else {
            SS.Print(PIF.PRINT_ERRORS(Cptr->is_cgi));
        }
        if (!Cptr->is_cgi) {
            SS.Print("There are RUN-TIME errors.", -1, -2);
        } else {
            SS.Print("\n\nThere are RUN-TIME errors.\n");
        }

        NotifyParent(pipe_out, parent, -2, PIF.SerializeWarningsErrors());
        return(0);
    }
    return 0;
}
//-----------------------------------------------------------------------------------
CONCEPT_DLL_API Concept_Execute3_Done(void *PTR) {
    Container *Cptr=(Container *)PTR;
    if (!Cptr)
        return -1;
    SimpleStream *SS=(SimpleStream *)Cptr->R2;
    PIFAlizator *PIF=(PIFAlizator *)Cptr->R1;

    delete Cptr;
    delete SS;
    delete PIF;
    return 0;
}
//-----------------------------------------------------------------------------------
#ifdef _WIN32
//-----------------------------------------------------------------------------------
// DLL Entry Point ... nothing ...
int WINAPI DllEntryPoint(HINSTANCE hinst, uintptr_t reason, void *lpReserved) {
    return(1);
}

//-----------------------------------------------------------------------------------
#endif
